@@ -69,17 +69,19 @@
 %   Translate a message Term into message lines. The produced lines
 %   is a list of
 %
-%       * nl
-%       Emit a newline
-%       * Fmt-Args
-%       Emit the result of format(Fmt, Args)
-%       * Fmt
-%       Emit the result of format(Fmt)
-%       * flush
-%       Used only as last element of the list.   Simply flush the
-%       output instead of producing a final newline.
-%       * at_same_line
-%       Start the messages at the same line (instead of using ~N)
+%       - nl
+%         Emit a newline
+%       - Fmt-Args
+%         Emit the result of format(Fmt, Args)
+%       - Fmt
+%         Emit the result of format(Fmt)
+%       - ansi(Code, Fmt, Args)
+%         Use ansi_format/3 for color output.
+%       - flush
+%         Used only as last element of the list.   Simply flush the
+%         output instead of producing a final newline.
+%       - at_same_line
+%         Start the messages at the same line (instead of using ~N)
 
 translate_message(Term) -->
     translate_message2(Term),
@@ -136,6 +138,8 @@ term_message(Term) -->
 term_message(Term) -->
     [ 'Unknown error term: ~p'-[Term] ].
 
+iso_message(resource_error(c_stack)) -->
+    out_of_c_stack.
 iso_message(resource_error(Missing)) -->
     [ 'Not enough resources: ~w'-[Missing] ].
 iso_message(type_error(evaluable, Actual)) -->
@@ -217,6 +221,12 @@ permission_error(open, source_sink, alias(Alias)) -->
     [ 'No permission to reuse alias "~p": already taken'-[Alias] ].
 permission_error(tnot, non_tabled_procedure, Pred) -->
     [ 'The argument of tnot/1 is not tabled: ~p'-[Pred] ].
+permission_error(assert, procedure, Pred) -->
+    { '$pi_head'(Pred, Head),
+      predicate_property(Head, ssu)
+    },
+    [ '~p: an SSU (Head => Body) predicate cannot have normal Prolog clauses'-
+      [Pred] ].
 permission_error(Action, Type, Object) -->
     [ 'No permission to ~w ~w `~p'''-[Action, Type, Object] ].
 
@@ -989,6 +999,29 @@ resolve_overflow(enlarge) -->
       '?- set_prolog_flag(stack_limit, ~I). to double the limit.'-[NewLimit]
     ].
 
+%!  out_of_c_stack
+%
+%   The thread's C-stack limit was exceeded. Give  some advice on how to
+%   resolve this.
+
+out_of_c_stack -->
+    { statistics(c_stack, Limit), Limit > 0 },
+    !,
+    [ 'C-stack limit (~D bytes) exceeded.'-[Limit], nl ],
+    resolve_c_stack_overflow(Limit).
+out_of_c_stack -->
+    { statistics(c_stack, Limit), Limit > 0 },
+    [ 'C-stack limit exceeded.'-[Limit], nl ],
+    resolve_c_stack_overflow(Limit).
+
+resolve_c_stack_overflow(_Limit) -->
+    { thread_self(main) },
+    [ 'Use the shell command ' ], code('~w', 'ulimit -s size'),
+    [ ' to enlarge the limit.' ].
+resolve_c_stack_overflow(_Limit) -->
+    [ 'Use the ' ], code('~w', 'c_stack(KBytes)'),
+    [ ' option of '], code(thread_create/3), [' to enlarge the limit.' ].
+
 
                  /*******************************
                  *        MAKE/AUTOLOAD         *
@@ -1842,6 +1875,12 @@ print_system_message(_, Kind, Lines) :-
     msg_property(Kind, stream(Stream)),
     print_message_lines(Stream, kind(Kind), Lines).
 
+to_list(ListIn, List) :-
+    is_list(ListIn),
+    !,
+    List = ListIn.
+to_list(NonList, [NonList]).
+
 :- multifile
     user:message_property/2.
 
@@ -2050,6 +2089,14 @@ actions_to_format([ansi(_Attrs, Fmt0, Args0)|Tail], Fmt, Args) :-
     actions_to_format(Tail, Fmt1, Args1),
     atom_concat(Fmt0, Fmt1, Fmt),
     append_args(Args0, Args1, Args).
+actions_to_format([url(Pos)|Tail], Fmt, Args) :-
+    !,
+    actions_to_format(Tail, Fmt1, Args1),
+    url_actions_to_format(url(Pos), Fmt1, Args1, Fmt, Args).
+actions_to_format([url(URL, Label)|Tail], Fmt, Args) :-
+    !,
+    actions_to_format(Tail, Fmt1, Args1),
+    url_actions_to_format(url(URL, Label), Fmt1, Args1, Fmt, Args).
 actions_to_format([Fmt0-Args0|Tail], Fmt, Args) :-
     !,
     actions_to_format(Tail, Fmt1, Args1),
@@ -2074,14 +2121,33 @@ action_skip(flush).
 action_skip(begin(_Level, _Ctx)).
 action_skip(end(_Ctx)).
 
+url_actions_to_format(url(File:Line:Column), Fmt1, Args1, Fmt, Args) :-
+    !,
+    atom_concat('~w:~d:~d', Fmt1, Fmt),
+    append_args([File,Line,Column], Args1, Args).
+url_actions_to_format(url(File:Line), Fmt1, Args1, Fmt, Args) :-
+    !,
+    atom_concat('~w:~d', Fmt1, Fmt),
+    append_args([File,Line], Args1, Args).
+url_actions_to_format(url(File), Fmt1, Args1, Fmt, Args) :-
+    !,
+    atom_concat('~w', Fmt1, Fmt),
+    append_args([File], Args1, Args).
+url_actions_to_format(url(_URL, Label), Fmt1, Args1, Fmt, Args) :-
+    !,
+    atom_concat('~w', Fmt1, Fmt),
+    append_args([Label], Args1, Args).
+
+
 append_args(M:Args0, Args1, M:Args) :-
     !,
     strip_module(Args1, _, A1),
-    '$append'(Args0, A1, Args).
+    to_list(Args0, Args01),
+    '$append'(Args01, A1, Args).
 append_args(Args0, Args1, Args) :-
     strip_module(Args1, _, A1),
-    '$append'(Args0, A1, Args).
-
+    to_list(Args0, Args01),
+    '$append'(Args01, A1, Args).
 
                  /*******************************
                  *    MESSAGES TO PRINT ONCE    *
