@@ -168,7 +168,7 @@ class PrologError(Exception):
 
 class PrologLaunchError(Exception):
     """
-    Raised when the SWI Prolog process was unable to be launched for any reason.
+    Raised when the SWI Prolog process was unable to be launched for any reason. This can include a version mismatch between the library and the server.
     """
 
     pass
@@ -351,7 +351,8 @@ class PrologMQI:
                     with self.create_thread() as prologThread:
                         prologThread.halt_server()
 
-                result = self._process.wait()
+                # Use __exit__ instead of wait() as it will close stderr/out handles too
+                self._process.__exit__(None, None, None)
 
             # Need to get rid of the unix domain socket file
             if self._unix_domain_socket:
@@ -530,6 +531,8 @@ class PrologThread:
         self.communication_thread_id = None
         self.goal_thread_id = None
         self._heartbeat_count = 0
+        self._server_protocol_major = None
+        self._server_protocol_minor = None
 
     def __enter__(self):
         self.start()
@@ -596,6 +599,32 @@ class PrologThread:
         else:
             threadTerm = prolog_args(jsonResult)[0][0][0]
             self.communication_thread_id, self.goal_thread_id = prolog_args(threadTerm)
+            if len(prolog_args(jsonResult)[0][0]) > 1:
+                versionTerm = prolog_args(jsonResult)[0][0][1]
+                self._server_protocol_major = prolog_args(versionTerm)[0]
+                self._server_protocol_minor = prolog_args(versionTerm)[1]
+            else:
+                self._server_protocol_major = 0
+                self._server_protocol_minor = 0
+
+        self._check_protocol_version()
+
+    # Major versions get incremented when there is a change that will break clients written to the old version
+    # Minor versions get incremented if there are changes but it will not break clients written to the previous version
+    def _check_protocol_version(self):
+        # This version of the library works around a protocol bug in MQI 0.0, so it supports that *and* MQI 1.x
+        required_server_major = 1
+        required_server_minor = 0
+
+        if self._server_protocol_major == 0 and self._server_protocol_minor == 0:
+            return
+
+        if self._server_protocol_major == required_server_major and self._server_protocol_minor >= required_server_minor:
+            return
+
+        raise PrologLaunchError(
+            f"This version of swiplserver requires MQI major version {required_server_major} and minor version >= {required_server_minor}. The server is running MQI '{self._server_protocol_major}.{self._server_protocol_minor}'."
+        )
 
     def stop(self):
         """
@@ -640,7 +669,7 @@ class PrologThread:
             list: The query succeeded once with free variables or more than once with no free variables. There will be an item in the list for every answer. Each item will be:
 
                 - `True` if there were no free variables
-                - A `dict` if there were free variables. Each key will be the name of a variable, each value will be the JSON representing the term it was unified with.
+                - A `dict` if there were free variables. Each key will be the name of a variable, each value will be the JSON representing the term it was unified with. Note that a special variable called `$residuals` will be added to each answer that has residual [variable constraints on it](https://www.swi-prolog.org/pldoc/man?section=attvar). This will contain a list of all the constraints on all the variables for that answer.
         """
         if self._socket is None:
             self.start()
@@ -754,7 +783,8 @@ class PrologThread:
             list: The query succeeded once with free variables or more than once with no free variables. There will be an item in the list for every answer. Each item will be:
 
                 - `True` if there were no free variables
-                - A `dict` if there were free variables. Each key will be the name of a variable, each value will be the JSON representing the term it was unified with.
+                - A `dict` if there were free variables. Each key will be the name of a variable, each value will be the JSON representing the term it was unified with. Note that a special variable called `$residuals` will be added to each answer that has residual [variable constraints on it](https://www.swi-prolog.org/pldoc/man?section=attvar). This will contain a list of all the constraints on all the variables for that answer.
+
         """
         timeoutString = (
             "-1" if wait_timeout_seconds is None else str(wait_timeout_seconds)
@@ -835,7 +865,14 @@ class PrologThread:
         value += ".\n"
         _log.debug("PrologMQI send: %s", value)
         utf8Value = value.encode("utf-8")
-        msgHeader = f"{str(len(utf8Value))}.\n".encode("utf-8")
+        if self._server_protocol_major == 0:
+            # Work around bug in version 0 of MQI that required
+            # count of Unicode code points (not bytes)
+            messageLen = len(value)
+        else:
+            messageLen = len(utf8Value)
+
+        msgHeader = f"{str(messageLen)}.\n".encode("utf-8")
         self._socket.sendall(msgHeader)
         self._socket.sendall(utf8Value)
 
@@ -994,13 +1031,24 @@ class _NonBlockingStreamReader:
     def __init__(self, stream):
         def _print_output(stream):
             while True:
-                line = stream.readline()
+                with suppress(Exception):
+                    line = stream.readline()
+
+                # When the process exits the stream will return EOF
+                # and allow us to exit the thread cleanly
                 if line:
                     _log.critical(f"Prolog: {line.decode().rstrip()}")
+                else:
+                    break
 
         self._stream = stream
         self._thread = Thread(target=_print_output, args=(self._stream,))
-        self._thread.daemon = True
+        # Don't run the thread as a daemon as it causes errors on exit like these:
+        # https://bugs.python.org/issue42717
+        # https://bugs.python.org/issue23309
+        # https: // bugs.python.org / issue43308
+
+        # self._thread.daemon = True
         self._thread.start()
 
 
