@@ -73,7 +73,7 @@
 #define PUBLIC_GLOBAL extern		/* exported global variables */
 #endif
 
-#if defined(WIN32) || defined(__CYGWIN__)
+#if defined(WIN32)
 #define __pce_export __declspec(dllexport) extern
 #else
 #define __pce_export extern
@@ -394,8 +394,8 @@ test, conversion and computation macro's are provided.
 #define INTPTR_FORMAT "%d"
 #elif SIZEOF_LONG == SIZEOF_VOIDP
 #define INTPTR_FORMAT "%ld"
-#elif defined(__WIN64)
-#define INTPTR_FORMAT "%I64d"
+#elif SIZEOF_LONG_LONG == SIZEOF_VOIDP
+#define INTPTR_FORMAT "%lld"
 #else
 #error "Config needs INTPTR_FORMAT"
 #endif
