@@ -3,8 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2011-2017, University of Amsterdam
+    Copyright (c)  2011-2021, University of Amsterdam
                               VU University Amsterdam
+			      CWI, Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -64,15 +66,19 @@ locking is required.
 #undef O_LARGEFILES
 #endif
 
+#ifdef HAVE_VISIBILITY_ATTRIBUTE
+#define PL_SO_EXPORT __attribute__((visibility("default")))
+#endif
+
 #define PL_KERNEL 1
 #define O_LOCALE 1
 #include <wchar.h>
 #define NEEDS_SWINSOCK
 #define PL_STREAM_IMPL
 #include "pl-stream.h"
-#define PL_ARITY_AS_SIZE
-#include "SWI-Prolog.h"
 #include "pl-utf8.h"
+#include "../pl-mutex.h"
+#include "SWI-Prolog.h"
 #include <sys/types.h>
 #ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
@@ -156,7 +162,6 @@ STRYLOCK(IOSTREAM *s)
 #include "os/pl-locale.h"
 #endif
 
-extern int			fatalError(const char *fm, ...);
 extern int			PL_handle_signals();
 extern IOENC			initEncoding(void);
 extern int			reportStreamError(IOSTREAM *s);
@@ -195,6 +200,7 @@ S__setbuf(IOSTREAM *s, char *buffer, size_t size)
   } else
   { if ( !(newunbuf = malloc(size+UNDO_SIZE)) )
     { errno = ENOMEM;
+      S__seterror(s);
       return -1;
     }
     newflags &= ~SIO_USERBUF;
@@ -230,7 +236,8 @@ S__setbuf(IOSTREAM *s, char *buffer, size_t size)
       }
     }
 
-    memcpy(newbuf, s->bufp, copy);
+    if ( copy > 0 )
+      memcpy(newbuf, s->bufp, copy);
     S__removebuf(s);
     s->unbuffer = newunbuf;
     s->bufp = s->buffer = newbuf;
@@ -439,7 +446,6 @@ S__wait(IOSTREAM *s)
     if ( rc < 0 && errno == EINTR )
     { if ( PL_handle_signals() < 0 )
       { Sset_exception(s, PL_exception(0));
-	s->flags |= SIO_CLEARERR;
 	errno = EPLEXCEPTION;
 	return -1;
       }
@@ -524,11 +530,10 @@ S__flushbuf(IOSTREAM *s)
     { if ( errno == EINTR )
       { if ( PL_handle_signals() < 0 )
 	{ Sset_exception(s, PL_exception(0));
-	  s->flags |= SIO_CLEARERR;
 	  errno = EPLEXCEPTION;
 	} else
 	  goto retry;
-      } else if ( errno != EPLEXCEPTION && !(s->flags&SIO_NOERROR) )
+      } else
 	S__seterror(s);
       rc = -1;
       goto out;
@@ -675,8 +680,7 @@ S__fillbuf(IOSTREAM *s)
 	return -1;
 #endif
       } else
-      { if ( !(s->flags & SIO_FERR) )	/* error already set */
-	  S__seterror(s);
+      {	S__seterror(s);
 	return -1;
       }
     }
@@ -720,14 +724,6 @@ update_linepos(IOSTREAM *s, int c)
 
 
 
-int
-S__fcheckpasteeof(IOSTREAM *s, int c)
-{ S__checkpasteeof(s, c);
-
-  return c;
-}
-
-
 int
 S__fupdatefilepos_getc(IOSTREAM *s, int c)
 { IOPOS *p = s->position;
@@ -750,7 +746,6 @@ S__updatefilepos(IOSTREAM *s, int c)
   { update_linepos(s, c);
     p->charno++;
   }
-  S__checkpasteeof(s,c);
 
   return c;
 }
@@ -840,17 +835,20 @@ Sungetc(int c, IOSTREAM *s)
 
 static int
 reperror(int c, IOSTREAM *s)
-{ if ( c >= 0 && (s->flags & (SIO_REPXML|SIO_REPPL)) )
+{ if ( c >= 0 && (s->flags & (SIO_REPXML|SIO_REPPL|SIO_REPPLU)) )
   { char buf[16];
     const char *q;
 
     if ( (s->flags & SIO_REPPL) )
+    { sprintf(buf, "\\x%X\\", c);
+    } else if ( (s->flags & SIO_REPPLU) )
     { if ( c <= 0xffff )
 	sprintf(buf, "\\u%04X", c);
       else
 	sprintf(buf, "\\U%08X", c);
     } else
-      sprintf(buf, "&#%d;", c);
+    { sprintf(buf, "&#%d;", c);
+    }
 
     for(q = buf; *q; q++)
     { if ( put_byte(*q, s) < 0 )
@@ -860,7 +858,7 @@ reperror(int c, IOSTREAM *s)
     return c;
   }
 
-  Sseterr(s, SIO_FERR|SIO_CLEARERR, "Encoding cannot represent character");
+  Sseterr(s, SIO_FERR, "Encoding cannot represent character");
   return -1;
 }
 
@@ -1350,7 +1348,7 @@ Sread_pending(IOSTREAM *s, char *buf, size_t limit, int flags)
   { int c = S__fillbuf(s);
 
     if ( c < 0 )
-    { if ( (s->flags & SIO_FEOF) )
+    { if ( (s->flags & SIO_FEOF) && Sfpasteof(s) != TRUE )
 	return 0;
       return c;
     }
@@ -1453,8 +1451,7 @@ ScheckBOM(IOSTREAM *s)
       return 0;
 
     if ( S__fillbuf(s) == -1 )
-    { if ( s->limitp - s->bufp > 0 )
-	s->flags &= ~SIO_FEOF;
+    { s->flags &= ~SIO_FEOF;
       return 0;				/* empty stream */
     }
     s->bufp--;
@@ -1508,7 +1505,10 @@ Sfeof(IOSTREAM *s)
 
 static int
 S__seterror(IOSTREAM *s)
-{ s->io_errno = errno;
+{ if ( (s->flags & SIO_FERR) )
+    return 0;				/* error already set */
+
+  s->io_errno = errno;
 
   if ( !(s->flags&SIO_CLOSING) &&	/* s->handle is already invalid */
        s->functions->control )
@@ -1548,7 +1548,7 @@ Sfpasteof(IOSTREAM *s)
 
 
 #define SIO_ERROR_FLAGS (SIO_FEOF|SIO_WARN|SIO_FERR| \
-			 SIO_FEOF2|SIO_TIMEOUT|SIO_CLEARERR)
+			 SIO_FEOF2|SIO_TIMEOUT)
 
 void
 Sclearerr(IOSTREAM *s)
@@ -1568,7 +1568,7 @@ Sclearerr(IOSTREAM *s)
 int
 Sseterr(IOSTREAM *s, int flags, const char *message)
 { for(; s && s->magic == SIO_MAGIC; s = s->upstream )
-  { s->flags = (s->flags & ~(SIO_WARN|SIO_FERR|SIO_CLEARERR)) | flags;
+  { s->flags = (s->flags & ~(SIO_WARN|SIO_FERR)) | flags;
 
     if ( s->message )
     { free(s->message);
@@ -2018,12 +2018,16 @@ out:
 char *
 Sgets(char *buf)
 { char *s = Sfgets(buf, Slinesize, Sinput);
-  char *q;
 
   if ( s )				/* delete trailing \n */
-  { q = &s[strlen(s)];
-    if ( q > s && q[-1] == '\n' )
-      *--q = '\0';
+  { char *q;
+
+    for(q=s; q<s+Slinesize; q++)
+    { if ( *q == '\n' )
+      { *q = '\0';
+	break;
+      }
+    }
   }
 
   return s;
@@ -2082,31 +2086,53 @@ Svprintf(const char *fm, va_list args)
 { return Svfprintf(Soutput, fm, args);
 }
 
-
-#define NEXTCHR(s, c)				\
-	switch (enc)				\
-	{ case ENC_ANSI:			\
-	  case ENC_ISO_LATIN_1:			\
-	    c = *(s)++; c &= 0xff;		\
-	    break;				\
-	  case ENC_UTF8:			\
-	    (s) = utf8_get_char((s), &(c));	\
-	    break;				\
-	  case ENC_WCHAR:			\
-	  { wchar_t *_w = (wchar_t*)(s);	\
-	    c = *_w++;				\
-	    (s) = (char*)_w;			\
-	    break;				\
-	  }					\
-	  default:				\
-	    c = 0;				\
-	    assert(0);				\
-	    break;				\
-	}
+static int
+next_chr(const char **s, IOENC enc)
+{ switch(enc)
+  { case ENC_ANSI:
+    case ENC_ISO_LATIN_1:
+    { unsigned char c = (unsigned char)**s;
+      ++(*s);
+      return c;
+    }
+    case ENC_UTF8:
+    { int c;
+      PL_utf8_code_point(s, NULL, &(c));
+      return c;
+    }
+    case ENC_WCHAR:
+    { const wchar_t *w = (const wchar_t*)*s;
+      wint_t c = *w++;
+      *s = (const char*)w;
+      return c;
+    }
+    default:
+      assert(0);
+      return -1;
+  }
+}
 
 #define OUTCHR(s, c)	do { printed++; \
 			     if ( Sputcode((c), (s)) < 0 ) goto error; \
 			   } while(0)
+#define OUTFS() \
+	do \
+	{ if ( fs == fbuf ) \
+	  { while(fs < fe) \
+	    { int c = next_chr((const char**)&fs, enc); \
+	      OUTCHR(s, c); \
+	    } \
+	  } else \
+	  { for(;;) \
+	    { int c = next_chr((const char**)&fs, enc); \
+	      if ( c ) \
+		OUTCHR(s, c); \
+	      else \
+		break; \
+	    } \
+	  } \
+	} while(0)
+
 #define valdigit(c)	((c) - '0')
 #define A_LEFT	0			/* left-aligned field */
 #define A_RIGHT 1			/* right-aligned field */
@@ -2236,8 +2262,11 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 
 	switch(*fm)
 	{ case 'c':
-	    *fe++ = va_arg(args, int);
+	  { int c = va_arg(args, int);
+	    fe = utf8_put_char(fe, c);
+	    enc = ENC_UTF8;
 	    break;
+	  }
 	  case 'p':
 	  { void *ptr = va_arg(args, void*);
 	    char fmbuf[8], *fp=fmbuf;
@@ -2344,18 +2373,20 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	    break;
 	}
 
+	/* Now `fs` is either the result of some sub-formatting or the
+           argument of `%s`.  In the latter case it is by definition
+	   0-terminated.  In the former case it may hold 0-bytes, e.g.,
+	   `%c` using a 0 argument.  `fe` points at the end.
+	*/
+
 	if ( has_arg1 )			/* aligned field */
 	{ if ( fs == fbuf )
 	    *fe = '\0';
 
 	  if ( align == A_LEFT )
-	  { int w = 0;
-	    while(*fs)
-	    { int c;
-	      NEXTCHR(fs, c);
-	      OUTCHR(s, c);
-	      w++;
-	    }
+	  { int printed0 = printed;
+	    OUTFS();
+	    int w = printed-printed0;
 	    while(w < arg1)
 	    { OUTCHR(s, pad);
 	      w++;
@@ -2391,23 +2422,10 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 		w--;
 	      }
 	    }
-	    while(*fs)
-	    { int c;
-	      NEXTCHR(fs, c);
-	      OUTCHR(s, c);
-	    }
+	    OUTFS();
 	  }
 	} else
-	{ if ( fs == fbuf )		/* unaligned field, just output */
-	  { while(fs < fe)
-	      OUTCHR(s, *fs++);
-	  } else
-	  { while(*fs)
-	    { int c;
-	      NEXTCHR(fs, c);
-	      OUTCHR(s, c);
-	    }
-	  }
+	{ OUTFS();
 	}
 	fm++;
 	if ( fs_malloced )
@@ -2523,6 +2541,9 @@ Svdprintf(const char *fm, va_list args)
   return rval;
 }
 
+#ifdef O_DEBUG
+#undef Sdprintf
+#endif
 
 int
 Sdprintf(const char *fm, ...)
@@ -2536,6 +2557,35 @@ Sdprintf(const char *fm, ...)
   return rval;
 }
 
+#ifdef O_DEBUG
+
+int
+Sdprintf_ex(const char *channel, const char *file, int line, const char *fm, ...)
+{ va_list args;
+  int rval;
+
+  if ( Serror->position && Serror->position->linepos == 0 && channel)
+  { const char *logfmt = "[%s] %s:%d: ";
+
+    if (strncmp(channel, "DBG_LEVEL", 9) == 0)
+    { channel += 9;
+      logfmt = "<%s> %s:%d: ";
+    }
+    if (strncmp(file, "../", 3) == 0)
+      file += 3;
+    Sdprintf(logfmt, channel, file, line);
+  }
+
+  va_start(args, fm);
+  rval = Svdprintf(fm, args);
+  va_end(args);
+
+  return rval;
+}
+
+#define Sdprintf(fmt...) Sdprintf_ex(NULL, __FILE__, __LINE__, fmt)
+#endif /*O_DEBUG*/
+
 #if 0
 		 /*******************************
 		 *	      SCANF		*
@@ -2981,11 +3031,54 @@ IOFUNCTIONS Sfilefunctions =
 };
 
 
+		 /*******************************
+		 *	    TTY STREAMS		*
+		 *******************************/
+
+static ssize_t
+Sread_tty(void *handle, char *buf, size_t size)
+{ ssize_t bytes = Sread_file(handle, buf, size);
+
+  if ( bytes < 0 && errno == EBADF )
+  { errno = 0;
+    return 0;
+  }
+
+  return bytes;
+}
+
+
+static ssize_t
+Swrite_tty(void *handle, char *buf, size_t size)
+{ ssize_t bytes = Swrite_file(handle, buf, size);
+
+  if ( bytes == -1 && errno == EBADF )
+  { errno = 0;
+    return size;
+  }
+
+  return bytes;
+}
+
+
+static int
+Sclose_tty(void *handle)
+{ int rc = Sclose_file(handle);
+
+  if ( rc == -1 && errno == EBADF )
+  { errno = 0;
+    return 0;
+  }
+
+  return rc;
+}
+
+
 IOFUNCTIONS Sttyfunctions =
-{ Sread_file,
-  Swrite_file,
-  NULL,
-  Sclose_file,
+{ Sread_tty,
+  Swrite_tty,
+  Sseek_file,
+  Sclose_tty,
   Scontrol_file,
 #ifdef O_LARGEFILES
   Sseek_file64
@@ -3253,6 +3346,36 @@ Sopen_file(const char *path, const char *how)
 }
 
 
+IOSTREAM *
+Sopen_iri_or_file(const char *path, const char *how)
+{ int sl;
+  IOSTREAM *s;
+
+  if ( (sl=file_name_is_iri(path)) )
+  { atom_t mname = how[0] == 'r' ? ATOM_read : ATOM_write;
+    const char *m;
+
+    if ( !iri_hook(path, IRI_OPEN, mname, 0, &s) )
+      return NULL;
+
+    for(m=&how[1]; *m; m++)
+    { if ( *m == 'b' )
+      { clear(s, SIO_TEXT);
+	s->encoding = ENC_OCTET;
+	s->newline  = SIO_NL_POSIX;
+      } else if ( *m == 'r' )
+      { clear(s, SIO_RECORDPOS);
+	s->position = NULL;
+      }
+    }
+  } else
+  { s = Sopen_file(path, how);
+  }
+
+  return s;
+}
+
+
 IOSTREAM *
 Sfdopen(int fd, const char *type)
 { intptr_t lfd;
@@ -3345,14 +3468,6 @@ Swinsock(IOSTREAM *s)
 		 *******************************/
 
 #ifdef HAVE_POPEN
-#ifdef __WINDOWS__
-#include "windows/popen.c"
-
-#undef popen
-#undef pclose
-#define popen(cmd, how) pt_popen(cmd, how)
-#define pclose(fd)	pt_pclose(fd)
-#endif
 
 static ssize_t
 Sread_pipe(void *handle, char *buf, size_t size)
@@ -3557,24 +3672,24 @@ Sread_memfile(void *handle, char *buf, size_t size)
 }
 
 
-static long
-Sseek_memfile(void *handle, long offset, int whence)
+static int64_t
+Sseek_memfile64(void *handle, int64_t offset, int whence)
 { memfile *mf = handle;
 
   switch(whence)
   { case SIO_SEEK_SET:
       break;
     case SIO_SEEK_CUR:
-      offset += (long)mf->here;		/* Win64: truncates */
+      offset += mf->here;
       break;
     case SIO_SEEK_END:
-      offset = (long)mf->size - offset;	/* Win64 */
+      offset = mf->size - offset;
       break;
     default:
       errno = EINVAL;
       return -1;
   }
-  if ( offset < 0 || offset > (long)mf->size )
+  if ( offset < 0 || offset > mf->size )
   { errno = EINVAL;
     return -1;
   }
@@ -3584,6 +3699,24 @@ Sseek_memfile(void *handle, long offset, int whence)
 }
 
 
+static long
+Sseek_memfile(void *handle, long offset, int whence)
+{ memfile *mf = handle;
+  size_t old = mf->here;
+
+  int64_t rval = Sseek_memfile64(handle, (int64_t)offset, whence);
+  long rc = (long)rval;
+
+  if ( rval == rc )
+  { return rc;
+  } else
+  { mf->here = old;
+    errno = EINVAL;
+    return -1;
+  }
+}
+
+
 static int
 Sclose_memfile(void *handle)
 { memfile *mf = handle;
@@ -3599,12 +3732,42 @@ Sclose_memfile(void *handle)
   return -1;
 }
 
+static int
+Scontrol_memfile(void *handle, int action, void *arg)
+{ memfile *mf = handle;
+
+  switch(action)
+  { case SIO_SETENCODING:
+      errno = EPERM;
+      return -1;
+    case SIO_FLUSHOUTPUT:
+      return 0;
+    case SIO_GETSIZE:
+    { int64_t *szp = arg;
+
+      *szp = mf->here;
+      return 0;
+    }
+    case SIO_GETREPOSITION:
+    { int *valp = arg;
+      *valp = TRUE;
+
+      return 0;
+    }
+    default:
+      errno = EINVAL;
+      return -1;
+  }
+}
+
 
 IOFUNCTIONS Smemfunctions =
 { Sread_memfile,
   Swrite_memfile,
   Sseek_memfile,
-  Sclose_memfile
+  Sclose_memfile,
+  Scontrol_memfile,
+  Sseek_memfile64
 };
 
 
@@ -3634,7 +3797,7 @@ Sopenmem(char **buffer, size_t *sizep, const char* mode)
     Mode is "r" or "w".  The mode "rF" calls PL_free(*buffer) at when
     closed.
 
-Note: Its is NOT allows to access   streams  created with this call from
+Note: Its is NOT allowed to access   streams created with this call from
 multiple threads. This is ok for all   usage inside Prolog itself (often
 through tellString()/toldString(). This call is   intented  to use write
 and other output predicates to create strings.
@@ -3645,6 +3808,7 @@ Sopenmem(char **bufp, size_t *sizep, const char *mode)
 { memfile *mf = malloc(sizeof(memfile));
   int flags = SIO_FBUF|SIO_RECORDPOS|SIO_NOMUTEX|SIO_TEXT;
   size_t size;
+  IOSTREAM *s;
 
   if ( !mf )
   { errno = ENOMEM;
@@ -3694,7 +3858,10 @@ Sopenmem(char **bufp, size_t *sizep, const char *mode)
   mf->sizep	= sizep;
   mf->here      = 0;
 
-  return Snew(mf, flags, &Smemfunctions);
+  if ( (s=Snew(mf, flags, &Smemfunctions)) )
+    s->newline = SIO_NL_POSIX;
+
+  return s;
 }
 
 		 /*******************************
@@ -3849,8 +4016,8 @@ Sclose_buffer(IOSTREAM *s)
 #define SIO_STDIO (SIO_FILE|SIO_STATIC|SIO_NOCLOSE|SIO_ISATTY|SIO_TEXT)
 #define STDIO_STREAMS \
   STDIO(0, SIO_STDIO|SIO_LBUF|SIO_INPUT|SIO_NOFEOF),	/* Sinput */ \
-  STDIO(1, SIO_STDIO|SIO_LBUF|SIO_OUTPUT|SIO_REPPL),	/* Soutput */ \
-  STDIO(2, SIO_STDIO|SIO_NBUF|SIO_OUTPUT|SIO_REPPL)	/* Serror */
+  STDIO(1, SIO_STDIO|SIO_LBUF|SIO_OUTPUT|SIO_REPPLU),	/* Soutput */ \
+  STDIO(2, SIO_STDIO|SIO_NBUF|SIO_OUTPUT|SIO_REPPLU)	/* Serror */
 
 
 IOSTREAM S__iob[] =
@@ -3892,10 +4059,6 @@ SinitStreams(void)
       s->newline = SIO_NL_DOS;
 #endif
     }
-
-#ifdef __WINDOWS__
-    pt_init();				/* init popen() issues */
-#endif
   }
 }
 
