@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1997-2020, University of Amsterdam
+    Copyright (c)  1997-2021, University of Amsterdam
                               VU University Amsterdam
 			      CWI, Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -109,8 +110,8 @@ struct PL_global_data
   } resources;
 
   struct
-  { sig_handler handlers[MAXSIGNAL];	/* How Prolog preceives signals */
-#ifdef HAVE_SIGNAL
+  { sig_handler handlers[NUM_SIGNALS];	/* How Prolog receives signals */
+#ifdef SIG_ALERT
     int		sig_alert;		/* our alert signal */
 #endif
   } signals;
@@ -451,11 +452,16 @@ struct PL_local_data
   struct PL_local_data *next_free;	/* see maybe_free_local_data() */
 
   struct
-  { int		pending[2];		/* PL_raise() pending signals */
+  { wsigmask_t	pending;		/* PL_raise() pending signals */
     int		current;		/* currently processing signal */
     int		is_sync;		/* current signal is synchronous */
 #ifndef __unix__
     int		forced;			/* Forced signal */
+#endif
+#ifdef O_C_STACK_GUARDED
+    jmp_buf	context;		/* Guarded for C-stack overflows */
+    int		sig_critical;		/* context is valid */
+    void       *alt_stack;		/* The stack */
 #endif
   } signal;
 
